import { Rect } from "scichart";

import { get_position_index } from "../utils/helpers";
import type { BasePlot } from "./baseplot";

export enum ELayoutMode {
    FixedGrid = "FixedGrid",
    DynamicGrid = "DynamicGrid",
    Custom = "Custom"
}

export abstract class Layout {
    public abstract apply_layout(plots: BasePlot[]): void;
}

class FixedGridLayout extends Layout {
    private rects: Rect[];

    constructor(n_columns: number, n_rows: number) {
        super();

        const width = 1 / n_columns;
        const height = 1 / n_rows;
        this.rects = Array(n_rows * n_columns).fill(new Rect(0, 0, 1, 1)).map((_, i) => {
            const { row_idx, col_idx } = get_position_index(i, n_columns);
            const top = row_idx / n_rows;
            const left = col_idx / n_columns;
            return new Rect(left, top, width, height);
        });
    }

    public apply_layout(plots: BasePlot[]): void {
        plots.forEach((plot, i) => {
            const rect = this.rects[i % this.rects.length];
            if (rect === undefined) { throw new Error(`Rect at index ${i} is undefined`); }
            plot.surface.subPosition = rect
        })
    }

}

class DynamicGridLayout extends Layout {
    public apply_layout(plots: BasePlot[]): void {
        const n_plots: number = plots.length;
        const n_columns = Math.ceil(Math.sqrt(n_plots));

        // coordinate system is [0, 1] x [0, 1]
        const width = 1 / n_columns; 
        const height = 1 / Math.ceil(n_plots / n_columns);

        plots.forEach((plot, i) => {
            const { row_idx, col_idx } = get_position_index(i, n_columns);
            const top = row_idx * height;
            const left = col_idx * width;
            const rect = new Rect(left, top, width, height);
            plot.surface.subPosition = rect;
        });
    }
}



export function layout_factory(mode: ELayoutMode): Layout {
    switch (mode) {
        case ELayoutMode.FixedGrid:
            return new FixedGridLayout(2, 2);
        case ELayoutMode.DynamicGrid:
            return new DynamicGridLayout();
        case ELayoutMode.Custom:
            throw new Error("Not implemented");
        default:
            throw new Error("Unknown layout mode: " + mode);
    }
}
